@@ -251,7 +251,7 @@ const ORIGINAL_MAX: u64 = 0;
 /// = 2048 * 2^(k-1)`, which is the k-1'th bucket's end. So, we would use the previous bucket
 /// for those lower values as it has better precision.
 ///
-#[derive(Debug)]
+#[derive(Debug, Clone)]
 pub struct Histogram<T: Counter> {
     auto_resize: bool,
 
@@ -462,6 +462,11 @@ impl<T: Counter> Histogram<T> {
     pub fn add<B: Borrow<Histogram<T>>>(&mut self, source: B) -> Result<(), AdditionError> {
         let source = source.borrow();
 
+        // If source is empty there's nothing to add
+        if source.is_empty() {
+            return Ok(());
+        }
+
         // make sure we can take the values in source
         let top = self.highest_equivalent(self.value_for(self.last_index()));
         if top < source.max() {
@@ -473,10 +478,17 @@ impl<T: Counter> Histogram<T> {
                 .map_err(|_| AdditionError::ResizeFailedUsizeTypeTooSmall)?;
         }
 
-        if self.bucket_count == source.bucket_count
+        let matching_buckets = self.bucket_count == source.bucket_count
             && self.sub_bucket_count == source.sub_bucket_count
-            && self.unit_magnitude == source.unit_magnitude
-        {
+            && self.unit_magnitude == source.unit_magnitude;
+        if matching_buckets && self.is_empty() {
+            // Counts arrays are of the same length and meaning.
+            // If self is empty (all counters are zeroes) we can copy the source histogram with a memory copy.
+            self.counts[..].copy_from_slice(&source.counts[..]);
+            self.total_count = source.total_count;
+            self.min_non_zero_value = source.min_non_zero_value;
+            self.max_value = source.max_value;
+        } else if matching_buckets {
             // Counts arrays are of the same length and meaning,
             // so we can just iterate and add directly:
             let mut observed_other_total_count: u64 = 0;
@@ -578,6 +590,11 @@ impl<T: Counter> Histogram<T> {
     ) -> Result<(), SubtractionError> {
         let subtrahend = subtrahend.borrow();
 
+        // If the source is empty there's nothing to subtract
+        if subtrahend.is_empty() {
+            return Ok(());
+        }
+
         // make sure we can take the values in source
         let top = self.highest_equivalent(self.value_for(self.last_index()));
         if top < self.highest_equivalent(subtrahend.max()) {
@@ -1378,9 +1395,27 @@ impl<T: Counter> Histogram<T> {
 
         let target_index = self.index_for_or_last(value);
         // TODO use RangeInclusive when it's stable to avoid checked_add
-        let total_to_current_index = (0..target_index.checked_add(1).expect("usize overflow"))
-            .map(|i| self.count_at_index(i).expect("index is <= last_index()"))
-            .fold(0_u64, |t, v| t.saturating_add(v.as_u64()));
+        let end = target_index.checked_add(1).expect("usize overflow");
+        // count the smaller half
+        let (slice, lower) = if target_index < self.counts.len() / 2 {
+            (&self.counts[0..end], true)
+        } else {
+            (&self.counts[end..], false)
+        };
+        let iter = slice.iter().map(Counter::as_u64);
+
+        let total_to_current_index = if self.total_count < u64::MAX {
+            // if the total didn't saturate then any partial count shouldn't either.
+            // iter::sum optimizes better than the saturating_add fallback below
+            iter.sum::<u64>()
+        } else {
+            iter.fold(0u64, u64::saturating_add)
+        };
+        let total_to_current_index = if lower {
+            total_to_current_index
+        } else {
+            self.total_count - total_to_current_index
+        };
         total_to_current_index.as_f64() / self.total_count as f64
     }
 
@@ -1748,14 +1783,6 @@ impl<T: Counter> RestatState<T> {
 // Trait implementations
 // ********************************************************************************************
 
-impl<T: Counter> Clone for Histogram<T> {
-    fn clone(&self) -> Self {
-        let mut h = Histogram::new_from(self);
-        h += self;
-        h
-    }
-}
-
 // make it more ergonomic to add and subtract histograms
 impl<'a, T: Counter> AddAssign<&'a Histogram<T>> for Histogram<T> {
     fn add_assign(&mut self, source: &'a Histogram<T>) {
